// lateral skin effect of tectangle busbar
// 2017-07-13, 2017-08-09 K.Hirabayashi

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <complex.h>

#define COMPLEX double complex
#define fmin(x, y)	((x < y) ? x : y)


COMPLEX busbar(double, double, double, double, double, double);
double R1(double, double);
double R2(double, double, double);
double R3(double, double, double, double, double, double);
int inverse(int, COMPLEX a[][*], double *, double *);

int main(int argc, char **argv)
{
  double a, b, s, f, g, u, w, rdc;
  COMPLEX z, z1;

  g = 1e8 / 2.83;	// conductivity
  u = 4.0 * M_PI * 1e-7;	// permeability
  b = 5e-3;		// thichness
  a = 20.0 * b;		// width
//a = 100.0 * b;

  rdc = 1.0 / (g * a * b);	// Rdc
  for (f = 1e0; f <= 1e3; f *= 1.05) {
	// isolated
	z = busbar(a, b, 0.0, u, g, f);
	printf("%g %g\n", f, creal(z));		// Rac/Rdc
//	printf("%g %g\n", f, -cimag(z) * rdc / (2.0 * M_PI * f));	// Li
	// return circuit
//	z1 = busbar(a, b, b * 1.0, u, g, f);
//	printf("%g %g %g\n", f, creal(z), creal(z1));	// Rac/Rdc
  }
}

// This programme calculates alternating current distriution in a rectangular
// conductor 2*a m x b m, by solving the approximation to the inhomogeneous
// integral equation of current density distributions.
// a = width / 2 (m)
// b = thickness (m)
// s = 0: isolated bus bar
// s > 0: distance of bus bar of return circuit (m)

COMPLEX busbar(double a, double b, double s, double u, double g, double f)
{
  int N, i, j;
  double e0, c, xi, xj, v, w, radian, i0, rdc, fmax;
  double detr, eps, av;
  COMPLEX zac, cur, z1;

  N = (int)(0.5 * a / b + 0.999);
  a = 0.5 * a / (double)N;

  COMPLEX z[N][N], x[N][N], curnt[N], volts[N];
  double size[N], angle[N];

  w = 2.0 * M_PI * f;
  rdc = 1.0 / (a * 2.0 * (double)N * b * g);
  fmax = 9.0 / (M_PI * g * u * b * b);	// 3.0 * skin-depth
  if (fmax < f)
	fprintf(stderr, "frequency must be less than %.2e Hz for this model.\n",
		f);
  e0 = 1.0;

  // construct x-matrix

  c = w * u * g * a * b / 2.0 / M_PI;
  for (i = 0; i < N; i++) {
	xi = ((double)i + 0.5) * a ;
	if (s == 0.0)
		x[i][i] = 1.0 + I * c * (R1(a, b) + R2(a, b, xi * 2.0));
	else
		x[i][i] = 1.0 + I * c * (R1(a, b) + R2(a, b, fabs(xi * 2.0))
			- R3(a, b, a, b, 0.0, s)
			- R3(a, b, a, b, xi * 2.0 - a, s));
	for (j = 0; j < N; j++) {
		if (j == i)
			continue;
		xj = ((double)j + 0.5) * a;
		if (s == 0.0)
			x[i][j] = I * c * (R2(a, b, fabs(xi - xj))
				+ R2(a, b, fabs(xi + xj)));
		else
			x[i][j] = I * c * (R2(a, b, fabs(xi - xj))
				+ R2(a, b, fabs(xi + xj))
				- R3(a, b, a, b, fabs(xi - xj) - a, s)
				- R3(a, b, a, b, fabs(xi + xj) - a, s));
	}
  }

  /* construct z-matrix */

  for (i = 0; i < N; i++)
	for (j = 0; j < N; j++)
		z[i][j] = x[i][j];

  // set up voltages and compute currents

  v = g * e0;
  for (i = 0; i < N; i++) {
	volts[i] = (COMPLEX)v;
  }

  inverse(N, z, &detr, &eps);

  for (i = 0; i < N; i++) {
	curnt[i] = 0.0;
	for (j = 0; j < N; j++)
		curnt[i] += z[j][i] * volts[j];
  }

  cur = (COMPLEX)0.0;
  for (i = 0; i < N; i++)
	cur += curnt[i];
  i0 = (double)N * v;	// DC current = N * g * e0
  zac = i0 / cur;

  return zac;
}

/*
 * Inverse(n, a, deltam eps)
 * This subroutine inverts the matrix a[][n] and returns it's
 * determinant delta.  eps is the smallest pivot used in the Gaussian
 * inversion.  The main program must dimension a[][n].
 */

int inverse(int n, COMPLEX a[][n], double *pd, double *pe)
{
  int i, j, k, ip;
  double delta, eps;
  COMPLEX y, w;

  COMPLEX c[n], d[n];
  int iz[n];

  delta = 1.0;
  for (j = 0; j < n; j++)
	iz[j] = j;
  for (i = 0; i < n; i++) {
    k = i;
    y = a[i][i];
    if (i < n - 1) {
	for (j = i + 1; j < n; j++) {
		w = a[i][j];
		if (cabs(w) > cabs(y)) {
			k = j;
			y = w;
		}
	}
    }
    delta *= y;
    if (i == 0)
	eps = cabs(y);
    eps = fmin(eps, cabs(y));
    if (eps == 0.0)
	return 0;
    y = 1.0 / y;
    for (j = 0; j < n; j++) {
	c[j] = a[j][k];
	a[j][k] = a[j][i];
	a[j][i] = -c[j] * y;
    }
    for (j = 0; j < n; j++) {
	a[i][j] *= y;
	d[j] = a[i][j];
    }
    a[i][i] = y;
    j = iz[i];
    iz[i] = iz[k];
    iz[k] = j;
    for (k = 0; k < n; k++) {
	if (k == i)
		continue;
	for (j = 0; j < n; j++)
		if (j != i)
			a[k][j] -= d[j] * c[k];
    }
  }
  for (i = 0; i < n; i++) {
g9:
    k = iz[i];
    if (k == i)
	continue;
    for (j = 0; j < n; j++) {
	w = a[i][j];
	a[i][j] = a[k][j];
	a[k][j] = w;
    }
    ip = iz[i];
    iz[i] = iz[k];
    iz[k] = ip;
    delta = -delta;
    goto g9;
  }
  *pd = delta;
  *pe = eps;

  return 0;
}

// log(GMD) of Rectangle conductor (GMD = Geometical Mean Distance)
//    a
// - +-+       +-+
// | | |       | |
// b | |       | |
// | | |       | |
// - +-+       +-+
//    |<---s--->|

// self GMD of rectangle conductor
// a = width, b = height

double R1(double a, double b)
{
  double a2, b2, ab, ab2, ba, ba2, x;

  a2 = a * a; b2 = b * b;
  ab = a / b; ba = b / a; ab2 = ab * ab; ba2 = ba * ba;

  x = 0.5 * log(a2 + b2) - ab2 / 12.0 * log(1.0 + ba2)
	- ba2 / 12.0 * log(ab2 + 1.0)
	+ 2.0 / 3.0 * (ab * atan2(b, a) + ba * atan2(a, b))
	- 25.0 / 12.0;
  return x;	// log(R)
} 

// mutual GMD of two adjucent rectangle con ductor
// a = width, b = height, s = distance, a <= s

double R2(double a, double b, double s)
{
  double a2, b2, b3, b4, s2, s3, s4, x;
  double spa, spa2, spa3, spa4, sma, sma2, sma3, sma4;

  if ((x = s - a) < -1e-9) {
	fprintf(stderr, "figure overlaps. (s = %g b = %g)\n", s, b);
	return NAN;
  }
  else if (x <= 1e-9)
	s = a * (1.0 + 1e-9);

  a2 = a * a; b2 = b * b; b3 = b2 * b; b4 = b3 * b;
  s2 = s * s; s3 = s2 * s; s4 = s3 * s;
  spa = s + a; spa2 = spa * spa; spa3 = spa2 * spa; spa4 = spa3 * spa;
  sma = s - a; sma2 = sma * sma; sma3 = sma2 * sma; sma4 = sma3 * sma;

  x =	  0.25 * (spa2 * (b2 - spa2 / 6.0) - b4 / 6.0) * log(spa2 + b2)
	+ 0.25 * (sma2 * (b2 - sma2 / 6.0) - b4 / 6.0) * log(sma2 + b2)
	- 0.5 * (s2 * (b2 - s2 / 6.0) - b4 / 6.0) * log(s2 + b2)
	+ spa4 * log(spa) / 12.0
	+ sma4 * log(sma) / 12.0
	- s4 * log(s) / 6.0
	+ (b * spa3 - b3 * spa) * atan2(b, spa) / 3.0
	+ (b * sma3 - b3 * sma) * atan2(b, sma) / 3.0
	- 2.0 / 3.0 * (b * s3 - b3 * s) * atan2(b, s) - 25.0 / 12.0 * a2 * b2;
  return x / a2 / b2;	// x = log(R)
}

//             R
//           +---+ -
//           |   | |
//           |   | S
//           |   | |
//           +---+ -
//     |--D--|     |
//     |           P
//    r|           |
// - +-+ -----------
// | | |
// s | |
// | | |
// - +-+

// D = P = 0.0 Ϸ׻Ǥʤ

double R3(double r, double s, double R, double S, double D, double P)
{
  int i, j;
  double A[4], B[4], d, r0, r1, r2, r3, r4;

  A[0] = fabs(D+ R + r);
  A[1] = fabs(D + R);
  A[2] = fabs(D);
  A[3] = fabs(D+ r);
  B[0] = fabs(P+ S + s);
  B[1] = fabs(P + S);
  B[2] = fabs(P);
  B[3] = fabs(P+ s);
  r0 = 0.0;
  for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
		if (A[i] == 0.0) {
			if (B[j] == 0.0)
				r4 = 0.0;
			else
				r4 = pow(B[j], 4.0) * log(B[j]);
		}
		else if (B[j] == 0.0)
			r4 = pow(A[i], 4.0) * log(A[i]);
		else {
			r1 = pow(A[i],4.0) - 6.0*pow(A[i]*B[j],2.0)
				+ pow(B[j],4.0);
			r1 *= log(hypot(A[i],B[j]));
			r2 = 4.0*A[i]*pow(B[j],3.0)*atan2(A[i],B[j]);
			r3 = 4.0*pow(A[i],3.0)*B[j]*atan2(B[j],A[i]);
			r4 = r1 - r2 - r3;
		}
		r0 += ((i + j) % 2) ? -r4 : r4;
	}
  }
  d = (-25.0/12.0*R*S*r*s - 1.0/24.0*r0) / (R*S*r*s);	// log(GMD)
  return d;
}
